@@ -16,7 +16,7 @@ from .errors import (
     check_call,
 )
 from .subscription import Subscription
-from .util import str2c
+from .util import c2str, str2c
 from .value import Value
 
 
@@ -137,7 +137,7 @@ class SysrepoSession:
     # end: general
 
     # begin: subscription
-    ModuleChangeCallbackType = Callable[[str, int, Dict, List[Change], Any], None]
+    ModuleChangeCallbackType = Callable[[str, int, List[Change], Any], None]
     """
     Callback to be called when the change in the datastore occurs.
 
@@ -147,11 +147,6 @@ class SysrepoSession:
     :arg req_id:
         Request ID unique for the specific module name. Connected events for one request
         ("change" and "done" for example) have the same request ID.
-    :arg config:
-        The configuration tree in a python dictionary. The contents are filtered
-        according to xpath (if set). If event is "update", "change" or "enabled", this
-        is the yet-to-be-applied configuration. If event is "done", this is the current
-        configuration that was just applied.
     :arg changes:
         List of `sysrepo.Change` objects representing what parts of the configuration
         have changed.
@@ -178,7 +173,8 @@ class SysrepoSession:
         done_only: bool = False,
         enabled: bool = False,
         private_data: Any = None,
-        asyncio_register: bool = False
+        asyncio_register: bool = False,
+        include_implicit_defaults: bool = True
     ) -> None:
         """
         Subscribe for changes made in the specified module.
@@ -212,12 +208,19 @@ class SysrepoSession:
         :arg asyncio_register:
             Add the created subscription event pipe into asyncio event loop
             monitored read file descriptors. Implies `no_thread=True`.
+        :arg include_implicit_defaults:
+            Include implicit default nodes in changes.
         """
         if self.is_implicit:
             raise SysrepoUnsupportedError("cannot subscribe with implicit sessions")
         _check_subscription_callback(callback, self.ModuleChangeCallbackType)
 
-        sub = Subscription(callback, private_data, asyncio_register=asyncio_register)
+        sub = Subscription(
+            callback,
+            private_data,
+            asyncio_register=asyncio_register,
+            include_implicit_defaults=include_implicit_defaults,
+        )
         sub_p = ffi.new("sr_subscription_ctx_t **")
 
         if asyncio_register:
@@ -321,19 +324,45 @@ class SysrepoSession:
 
         self.subscriptions.append(sub)
 
-    RpcCallbackType = Callable[[Dict, str, Any], Optional[Dict]]
+    RpcCallbackType = Callable[[str, Dict, str, Any], Optional[Dict]]
     """
     Callback to be called when the RPC/action is invoked.
 
-    :arg rpc_input:
-        The RPC input arguments in a python dictionary. If there are any input
-        parameters, the first level of the dictionary will have only one key with the
-        name of the RPC::
+    :arg xpath:
+        The full data path to the invoked RPC/action. When it is an RPC, the form is
+        `/prefix:rpc-name`. When it is an action, it is the full data path with all
+        parent nodes: `/prefix:container/list[key="val"]/action-name`.
+    :arg input_params:
+        The input arguments in a python dictionary. The contents are limited to the
+        children of the "input" node. For example, with a YANG rpc defined like this::
 
-            {'rpc-name': {'param1': 42, 'param2': 'foobar'}}
+            rpc rpc-name {
+              input {
+                leaf param1 {
+                  type uint32;
+                }
+                leaf param2 {
+                  type string;
+                }
+              }
+              output {
+                leaf foo {
+                  type int8;
+                }
+                leaf bar {
+                  type string;
+                }
+              }
+            }
 
-        If there are no input parameters provided by the client, the dictionary will be
-        empty: ``{}``.
+        The input_params dict may look like this::
+
+            {'param1': 42, 'param2': 'foobar'}
+
+        If there are no input parameters provided by the client, the dict will be empty.
+
+        For actions, the xpath argument allows to determine the parent node of the
+        action input parameters.
     :arg event:
         In most cases, it is always 'rpc'. When multiple callbacks are registered for
         the same RPC and one of the callbacks failed. The remainder of the callbacks
@@ -342,11 +371,16 @@ class SysrepoSession:
         Private context opaque to sysrepo used when subscribing.
 
     The callback is expected to return a python dictionary containing the RPC output
-    data. The dictionary should be in the libyang "dict" format. It will be passed to
-    libyang.DRpc.merge_data_dict() to return output data to sysrepo. If the callback
-    returns None, nothing is returned to sysrepo. If the callback raises an exception,
-    the error message is forwarded to the client that called the RPC. If the exception
-    is a subclass of SysrepoError, no traceback is sent to the logging system.
+    data. The dictionary should be in the libyang "dict" format and must only contain
+    the actual output parameters without any parents. Using the previous example::
+
+        {'foo': 47, 'bar': 'baz'}
+
+    It will be passed to libyang.DRpc.merge_data_dict() to return output data to
+    sysrepo. If the callback returns None, nothing is returned to sysrepo. If the
+    callback raises an exception, the error message is forwarded to the client that
+    called the RPC. If the exception is a subclass of SysrepoError, no traceback is sent
+    to the logging system.
     """
 
     def subscribe_rpc_call(
@@ -358,7 +392,8 @@ class SysrepoSession:
         no_thread: bool = False,
         private_data: Any = None,
         asyncio_register: bool = False,
-        strict: bool = False
+        strict: bool = False,
+        include_implicit_defaults: bool = True
     ) -> None:
         """
         Subscribe for the delivery of an RPC/action.
@@ -381,13 +416,19 @@ class SysrepoSession:
         :arg strict:
             Reject the whole data returned by callback if it contains elements without
             schema definition.
+        :arg include_implicit_defaults:
+            Include implicit defaults into input parameters passed to callbacks.
         """
         if self.is_implicit:
             raise SysrepoUnsupportedError("cannot subscribe with implicit sessions")
         _check_subscription_callback(callback, self.RpcCallbackType)
 
         sub = Subscription(
-            callback, private_data, asyncio_register=asyncio_register, strict=strict
+            callback,
+            private_data,
+            asyncio_register=asyncio_register,
+            strict=strict,
+            include_implicit_defaults=include_implicit_defaults,
         )
         sub_p = ffi.new("sr_subscription_ctx_t **")
 
@@ -412,7 +453,9 @@ class SysrepoSession:
     # end: subscription
 
     # begin: changes
-    def get_changes(self, xpath: str) -> Iterator[Change]:
+    def get_changes(
+        self, xpath: str, include_implicit_defaults: bool = True
+    ) -> Iterator[Change]:
         """
         Return an iterator that will yield all pending changes in the current session.
 
@@ -420,6 +463,8 @@ class SysrepoSession:
             Xpath selecting the requested changes. Be careful, you must select all the
             changes, not just subtrees! To get a full change subtree `//.` can be
             appended to the XPath.
+        :arg include_implicit_defaults:
+            Include implicit default nodes.
 
         :returns:
             An iterator that will yield `sysrepo.Change` objects.
@@ -429,30 +474,45 @@ class SysrepoSession:
         check_call(lib.sr_get_changes_iter, self.cdata, str2c(xpath), iter_p)
 
         op_p = ffi.new("sr_change_oper_t *")
-        old_p = ffi.new("sr_val_t **")
-        new_p = ffi.new("sr_val_t **")
+        node_p = ffi.new("struct lyd_node **")
+        prev_val_p = ffi.new("char **")
+        prev_list_p = ffi.new("char **")
+        prev_dflt_p = ffi.new("bool *")
+        ctx = self.get_ly_ctx()
 
         try:
             ret = check_call(
-                lib.sr_get_change_next,
+                lib.sr_get_change_tree_next,
                 self.cdata,
                 iter_p[0],
                 op_p,
-                old_p,
-                new_p,
+                node_p,
+                prev_val_p,
+                prev_list_p,
+                prev_dflt_p,
                 valid_codes=(lib.SR_ERR_OK, lib.SR_ERR_NOT_FOUND),
             )
             while ret == lib.SR_ERR_OK:
-                yield Change.parse(op_p[0], old=old_p[0], new=new_p[0])
-                lib.sr_free_val(old_p[0])
-                lib.sr_free_val(new_p[0])
+                try:
+                    yield Change.parse(
+                        operation=op_p[0],
+                        node=libyang.DNode.new(ctx, node_p[0]),
+                        prev_val=c2str(prev_val_p[0]),
+                        prev_list=c2str(prev_list_p[0]),
+                        prev_dflt=bool(prev_dflt_p[0]),
+                        include_implicit_defaults=include_implicit_defaults,
+                    )
+                except Change.Skip:
+                    pass
                 ret = check_call(
-                    lib.sr_get_change_next,
+                    lib.sr_get_change_tree_next,
                     self.cdata,
                     iter_p[0],
                     op_p,
-                    old_p,
-                    new_p,
+                    node_p,
+                    prev_val_p,
+                    prev_list_p,
+                    prev_dflt_p,
                     valid_codes=(lib.SR_ERR_OK, lib.SR_ERR_NOT_FOUND),
                 )
         finally:
